﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace Xenon.KifuNarabe
{

    /// <summary>
    /// 盤上用の升目符号。
    /// 
    /// 「先後」、
    /// 「置き場」、
    /// 「筋、段」（明記がない場合、dstの意）、
    /// 「src筋、src段」（オプション）。
    /// 
    /// ・“同”
    /// 棋譜用の升目符号。“同”が使えます。
    /// “同”にする場合は、UNKNOWN_SUJI 筋、UNKNOWN_DAN 段に設定してください。毎回再計算します。
    /// 
    /// ・前の升：「7g7f」などにも使えるよう、src筋、src段も補助で用意。
    /// ・前の手：また、遡れるように、１つ前の升目符号へのリンクも補助で容易。
    /// 
    /// ・駒種類：「歩」「と金」など。補助で容易。
    /// </summary>
    [Serializable]
    public class MasuBango
    {

        #region 定数
        //------------------------------------------------------------
        /// <summary>
        /// 初期設定のまま変わっていません。8～-8 程度だと、角等の射程に入るので、大きく外した数字をフラグに使います。
        /// </summary>
        public const int INITIALIZED_SUJI = -98;

        /// <summary>
        /// 初期設定のまま変わっていません。8～-8 程度だと、角等の射程に入るので、大きく外した数字をフラグに使います。
        /// </summary>
        public const int INITIALIZED_DAN = -98;

        /// <summary>
        /// 駒袋に入っていたとき。8～-8 程度だと、角等の射程に入るので、大きく外した数字をフラグに使います。
        /// </summary>
        public const int KOMABUKURO_SUJI = -97;

        /// <summary>
        /// 駒袋に入っていたとき。8～-8 程度だと、角等の射程に入るので、大きく外した数字をフラグに使います。
        /// </summary>
        public const int KOMABUKURO_DAN = -97;

        /// <summary>
        /// ボタンなので場所固定。8～-8 程度だと、角等の射程に入るので、大きく外した数字をフラグに使います。
        /// </summary>
        public const int BUTTON_SUJI = -96;

        /// <summary>
        /// ボタンなので場所固定。8～-8 程度だと、角等の射程に入るので、大きく外した数字をフラグに使います。
        /// </summary>
        public const int BUTTON_DAN = -96;

        /// <summary>
        /// 正しくない数字が入っているときです。8～-8 程度だと、角等の射程に入るので、大きく外した数字をフラグに使います。
        /// </summary>
        public const int ERROR_SUJI = -95;

        /// <summary>
        /// 正しくない数字が入っているときです。8～-8 程度だと、角等の射程に入るので、大きく外した数字をフラグに使います。
        /// </summary>
        public const int ERROR_DAN = -95;

        /// <summary>
        /// 一手前局面の駒位置を見る必要があります。8～-8 程度だと、角等の射程に入るので、大きく外した数字をフラグに使います。
        /// </summary>
        public const int ITTEMAE_KYOKUMEN_SUJI = -94;

        /// <summary>
        /// 一手前局面の駒位置を見る必要があります。8～-8 程度だと、角等の射程に入るので、大きく外した数字をフラグに使います。
        /// </summary>
        public const int ITTEMAE_KYOKUMEN_DAN = -94;

        /// <summary>
        /// 使わないはずの、srcのsrcです。8～-8 程度だと、角等の射程に入るので、大きく外した数字をフラグに使います。
        /// </summary>
        public const int TUKAWANAI_SRCSRC_SUJI = -93;

        /// <summary>
        /// 使わないはずの、srcのsrcです。8～-8 程度だと、角等の射程に入るので、大きく外した数字をフラグに使います。
        /// </summary>
        public const int TUKAWANAI_SRCSRC_DAN = -93;

        /// <summary>
        /// 符号で探し出せなかった駒の場合です。8～-8 程度だと、角等の射程に入るので、大きく外した数字をフラグに使います。
        /// </summary>
        public const int DA_SUJI = -92;

        /// <summary>
        /// 符号で探し出せなかった駒の場合です。8～-8 程度だと、角等の射程に入るので、大きく外した数字をフラグに使います。
        /// </summary>
        public const int DA_DAN = -92;
        //------------------------------------------------------------
        /// <summary>
        /// 筋は 1～9 だけ有効です。
        /// </summary>
        public const int YUKO_SUJI_MIN = 1;
        public const int YUKO_SUJI_MAX = 9;

        /// <summary>
        /// 段は 1～9 だけ有効です。
        /// </summary>
        public const int YUKO_DAN_MIN = 1;
        public const int YUKO_DAN_MAX = 9;
        //------------------------------------------------------------
        #endregion



        #region プロパティー類
        //------------------------------------------------------------
        private Sengo sengo;

        /// <summary>
        /// 先後
        /// </summary>
        public Sengo Sengo
        {
            get
            {
                return this.sengo;
            }
        }
        //------------------------------------------------------------
        private Okiba okiba;

        /// <summary>
        /// 駒の置き場
        /// </summary>
        public Okiba Okiba
        {
            get
            {
                return this.okiba;
            }
        }

        private Okiba srcOkiba;

        /// <summary>
        /// 駒の置き場
        /// </summary>
        public Okiba SrcOkiba
        {
            get
            {
                return this.srcOkiba;
            }
        }
        //------------------------------------------------------------
        private int suji;

        /// <summary>
        /// 筋。“同”解決付き
        /// </summary>
        public virtual int Suji
        {
            get
            {
                return this.suji;
            }
        }

        private int dan;

        /// <summary>
        /// 段。“同”解決付き
        /// </summary>
        public int Dan
        {
            get
            {
                return this.dan;
            }
        }
        //------------------------------------------------------------
        private int srcSuji;

        /// <summary>
        /// 移動前の筋
        /// </summary>
        public virtual int SrcSuji
        {
            get
            {
                return this.srcSuji;
            }
        }

        private int srcDan;

        /// <summary>
        /// 移動前の段
        /// </summary>
        public virtual int SrcDan
        {
            get
            {
                return this.srcDan;
            }
        }
        //------------------------------------------------------------
        private KomaSyurui syurui;
        /// <summary>
        /// 駒の種類。歩、桂…。
        /// </summary>
        public KomaSyurui Syurui
        {
            get
            {
                return this.syurui;
            }
        }
        //------------------------------------------------------------
        private KomaSyurui srcSyurui;
        /// <summary>
        /// 移動前の駒の種類。歩、桂…。  ＜不成判定用＞
        /// </summary>
        public KomaSyurui SrcSyurui
        {
            get
            {
                return this.srcSyurui;
            }
        }
        //------------------------------------------------------------
        #endregion

        public MasuBango(Sengo sengo, Okiba okiba, Okiba srcOkiba, MasuBango dst)
        {
            this.sengo = sengo;
            this.okiba = okiba;
            this.srcOkiba = srcOkiba;// Okiba.SHOGIBAN;
            this.suji = dst.Suji;
            this.dan = dst.Dan;
            this.srcSuji = 0;
            this.srcDan = 0;
            this.syurui = dst.Syurui;
            this.srcSyurui = dst.SrcSyurui;
        }

        /// <summary>
        /// 全部。
        /// </summary>
        /// <param name="sengo"></param>
        /// <param name="okiba"></param>
        /// <param name="srcOkiba"></param>
        /// <param name="suji"></param>
        /// <param name="dan"></param>
        /// <param name="srcSuji"></param>
        /// <param name="srcDan"></param>
        /// <param name="syurui"></param>
        /// <param name="srcSyurui"></param>
        /// <param name="previousTe"></param>
        public MasuBango(Sengo sengo, Okiba okiba, Okiba srcOkiba, int suji, int dan, int srcSuji, int srcDan, KomaSyurui syurui, KomaSyurui srcSyurui)
        {
            this.sengo = sengo;
            this.okiba = okiba;
            this.srcOkiba = srcOkiba;
            this.suji = suji;
            this.dan = dan;
            this.srcSuji = srcSuji;
            this.srcDan = srcDan;
            this.srcSyurui = srcSyurui;
            this.syurui = syurui;
        }


        public MasuBango Src()
        {
            MasuBango src = new MasuBango(
                this.Sengo,
                this.Okiba,
                this.SrcOkiba,
                this.srcSuji,
                this.srcDan,
                MasuBango.TUKAWANAI_SRCSRC_SUJI,
                MasuBango.TUKAWANAI_SRCSRC_DAN,
                this.SrcSyurui,
                KomaSyurui.FU//TODO:不明
                );

            return src;
        }


        #region 比較判定系
        //------------------------------------------------------------
        public bool MatchOkiba(Okiba okiba2)
        {
            return
                this.okiba == okiba2;
        }

        public bool MatchOkiba(MasuBango m2)
        {
            return this.MatchOkiba(m2.Okiba);
        }

        public bool MatchSengo(MasuBango m2)
        {
            return
                this.sengo == m2.sengo;
        }

        public bool MatchSuji(int suji2)
        {
            return
                // 有効な値かどうか
                MasuBango.YUKO_SUJI_MIN<=this.suji
                && this.suji<=MasuBango.YUKO_SUJI_MAX

                && this.suji == suji2;
        }

        public bool MatchSuji(MasuBango m2)
        {
            return this.MatchSuji(m2.Suji);
        }

        public bool MatchDan(int dan2)
        {
            return
                // 有効な値かどうか
                MasuBango.YUKO_DAN_MIN <= this.dan
                && this.dan <= MasuBango.YUKO_DAN_MAX

                && this.dan == dan2;
        }

        public bool MatchDan(MasuBango m2)
        {
            return this.MatchDan(m2.Dan);
        }
        //------------------------------------------------------------
        public bool MatchMasu(Object obj2)
        {
            bool result = false;

            if(obj2 is MasuBango)
            {
                MasuBango m2 = (MasuBango)obj2;
                result = this.MatchSuji(m2) && this.MatchDan(m2);
            }

            return result;
        }
        //------------------------------------------------------------
        public bool MatchKoma(Object obj2)
        {
            bool result = false;

            if (obj2 is MasuBango)
            {
                MasuBango m2 = (MasuBango)obj2;
                result = this.MatchSengo(m2)
                    && this.MatchSuji(m2)
                    && this.MatchDan(m2);
            }

            return result;
        }
        //------------------------------------------------------------
        /// <summary>
        /// 一致判定：   升目符号同士
        /// </summary>
        /// <param name="masu2"></param>
        /// <returns></returns>
        public bool Match(MasuBango masu2)
        {
            return
                this.MatchOkiba(masu2)
                && this.MatchSuji(masu2)
                && this.MatchDan(masu2);
        }
        //------------------------------------------------------------
        public bool ExistsIn(List<MasuBango> masu2Arr)
        {
            bool matched = false;

            foreach (MasuBango m in masu2Arr)
            {
                if (
                    null != m
                    && this.MatchSengo(m) //2014-07-21 先後も見るように追記。
                    && this.MatchSuji(m)
                    && this.MatchDan(m)
                    )
                {
                    matched = true;
                    break;
                }
            }

            return matched;
        }
        //------------------------------------------------------------
        #endregion



        /// <summary>
        /// １マス上、のように指定して、マスを取得します。
        /// </summary>
        /// <param name="offsetSuji"></param>
        /// <param name="offsetDan"></param>
        /// <returns></returns>
        public MasuBango Offset(Sengo sengo, int offsetSuji, int offsetDan)
        {
            return new MasuBango(
                sengo, 
                this.Okiba,
                this.SrcOkiba,
                this.Suji + offsetSuji,
                this.Dan + offsetDan,
                this.SrcSuji,
                this.SrcDan,
                this.Syurui,
                this.SrcSyurui
                );
        }


        public string ToText()
        {
            StringBuilder sb = new StringBuilder();


            //------------------------------------------------------------
            // 元筋
            //------------------------------------------------------------
            string strSrcSuji;
            switch (this.SrcSuji)
            {
                case BUTTON_SUJI:
                    strSrcSuji = "ボ";
                    break;
                case ERROR_SUJI:
                    strSrcSuji = "エ";
                    break;
                case ITTEMAE_KYOKUMEN_SUJI:
                    strSrcSuji = "前";
                    break;
                case KOMABUKURO_SUJI:
                    strSrcSuji = "袋";
                    break;
                case DA_SUJI:
                    strSrcSuji = Converter.SyuruiToSfenDa(this.Syurui);//打のとき。
                    break;
                case TUKAWANAI_SRCSRC_SUJI:
                    strSrcSuji = "ソ";
                    break;
                case INITIALIZED_SUJI:
                    strSrcSuji = "Ｉ";
                    break;
                default:
                    strSrcSuji = this.SrcSuji.ToString();
                    break;
            }
            sb.Append(strSrcSuji);

            //------------------------------------------------------------
            // 元段
            //------------------------------------------------------------
            string strSrcDan;
            switch (this.SrcDan)
            {
                case BUTTON_DAN:
                    strSrcDan = "ボ";
                    break;
                case ERROR_DAN:
                    strSrcDan = "エ";
                    break;
                case ITTEMAE_KYOKUMEN_DAN:
                    strSrcDan = "前";
                    break;
                case KOMABUKURO_DAN:
                    strSrcDan = "袋";
                    break;
                case DA_DAN:
                    strSrcDan = "*";//仕様：打のとき、元段は *表記 になります。
                    break;
                case TUKAWANAI_SRCSRC_DAN:
                    strSrcDan = "ソ";
                    break;
                case INITIALIZED_DAN:
                    strSrcDan = "Ｕ";
                    break;
                default:
                    strSrcDan = Converter.IntToAlphabet(this.SrcDan);
                    break;
            }
            sb.Append(strSrcDan);

            //------------------------------------------------------------
            // 現筋
            //------------------------------------------------------------
            string strSuji;
            switch (this.Suji)
            {
                case BUTTON_SUJI:
                    strSuji = "ボ";
                    break;
                case ERROR_SUJI:
                    strSuji = "エ";
                    break;
                case ITTEMAE_KYOKUMEN_SUJI:
                    strSuji = "前";
                    break;
                case KOMABUKURO_SUJI:
                    strSuji = "袋";
                    break;
                case DA_SUJI:
                    strSuji = "Ｄ";//エラー
                    break;
                case TUKAWANAI_SRCSRC_SUJI:
                    strSuji = "ソ";
                    break;
                case INITIALIZED_SUJI:
                    strSuji = "Ｉ";
                    break;
                default:
                    strSuji = this.Suji.ToString();
                    break;
            }
            sb.Append(strSuji);


            //------------------------------------------------------------
            // 現段
            //------------------------------------------------------------
            string strDan;
            switch (this.Dan)
            {
                case BUTTON_DAN:
                    strDan = "ボ";
                    break;
                case ERROR_DAN:
                    strDan = "エ";
                    break;
                case ITTEMAE_KYOKUMEN_DAN:
                    strDan = "前";
                    break;
                case KOMABUKURO_DAN:
                    strDan = "袋";
                    break;
                case DA_DAN:
                    strDan = "Ｄ";//エラー
                    break;
                case TUKAWANAI_SRCSRC_DAN:
                    strDan = "ソ";
                    break;
                case INITIALIZED_DAN:
                    strDan = "Ｕ";
                    break;
                default:
                    strDan = Converter.IntToAlphabet(this.Dan);
                    break;
            }
            sb.Append(strDan);


            //TODO:打
            if (this.IsDaAction)
            {
                sb.Append("+");
            }

            return sb.ToString();
        }

        public override string ToString()
        {
            return this.ToText();
        }


        /// <summary>
        /// 駒の移動。　前回の先を元へ、今回を先へ。
        /// </summary>
        /// <param name="okiba"></param>
        /// <param name="dst">移動先</param>
        public void MoveTo(MasuBango dst)
        {
            //------------------------------
            // 筋
            //------------------------------
            this.srcSuji = this.Suji;
            this.suji = dst.Suji;

            //------------------------------
            // 段
            //------------------------------
            this.srcDan = this.Dan;
            this.dan = dst.Dan;

            //------------------------------
            // 駒の種類
            //------------------------------
            //移動前の駒の種類
            KomaSyurui srcSyurui2;
            if (KomaSyurui.UNKNOWN == dst.SrcSyurui)
            {
                srcSyurui2 = this.SrcSyurui;
            }
            else
            {
                srcSyurui2 = dst.SrcSyurui;
            }
            this.srcSyurui = srcSyurui2;//移動前の駒の種類

            // 移動後の駒の種類
            KomaSyurui syurui2;
            if (KomaSyurui.UNKNOWN == dst.Syurui)
            {
                syurui2 = this.Syurui;
            }
            else
            {
                syurui2 = dst.Syurui;
            }
            this.syurui = syurui2;// 移動後の駒の種類


            //------------------------------
            // 先後
            //------------------------------
            this.sengo = dst.Sengo; //移動後の先後

            //------------------------------
            // 置き場
            //------------------------------
            this.srcOkiba = this.okiba;
            this.okiba = dst.Okiba; //移動後の置き場
        }


        /// <summary>
        /// 移動前と、移動後の場所が異なっていれば真。
        /// </summary>
        /// <returns></returns>
        public bool DoneMove()
        {
            return !this.Match(this.Src());
        }

        /// <summary>
        /// 不成ケース
        /// </summary>
        /// <returns></returns>
        public KomaSyurui ToFunariCase()
        {
            return Converter.SyuruiToFunariCase(this.Syurui);
        }

        /// <summary>
        /// 成ケース
        /// </summary>
        /// <returns></returns>
        public KomaSyurui ToNariCase()
        {
            return Converter.SyuruiToNariCase(this.Syurui);
        }

        /// <summary>
        /// 駒の表面の文字。
        /// </summary>
        public string Text_Label
        {
            get
            {
                return Converter.SyuruiToLabel(this.Syurui);
            }
        }

        /// <summary>
        /// 移動前の成り
        /// </summary>
        public bool PreNari
        {
            get
            {
                return Converter.SyuruiToNari(this.SrcSyurui);
            }
        }

        /// <summary>
        /// 成り
        /// </summary>
        public bool Nari
        {
            get
            {
                return Converter.SyuruiToNari(this.Syurui);
            }
        }

        /// <summary>
        /// “打” ＜アクション時＞
        /// </summary>
        /// <returns></returns>
        public bool IsDaAction
        {
            get
            {
                return Okiba.SHOGIBAN != this.SrcOkiba;// this.Src().Okiba;
            }
        }

        /// <summary>
        /// 一致判定：　先後、駒種類
        /// </summary>
        /// <returns></returns>
        public bool OnajiKoma(Button_Koma koma2)
        {
            return koma2 != null
                && this.Sengo == koma2.Masu.Sengo
                && this.Syurui == koma2.Masu.Syurui;
        }

        /// <summary>
        /// 不一致判定：　先後、駒種類  が、自分と同じものが　＜ひとつもない＞
        /// </summary>
        /// <returns></returns>
        public bool NeverOnaji(params Button_Koma[][] kGroupArgs)
        {
            bool unmatched = true;

            foreach (Button_Koma[] kGroup in kGroupArgs)
            {
                foreach (Button_Koma k in kGroup)
                {

                    if (
                        null != k
                        && this.Sengo == k.Masu.Sengo
                        && this.Syurui == k.Masu.Syurui
                        )
                    {
                        // １つでも一致するものがあれば、終了します。
                        unmatched = false;
                        goto gt_EndLoop;
                    }
                }

            }
        gt_EndLoop:

            return unmatched;
        }

        /// <summary>
        /// 一致判定：　先後、駒種類  が、自分と同じものが　＜ひとつ以上ある＞
        /// </summary>
        /// <returns></returns>
        public bool ExistsOnaji(params Button_Koma[][] kGroupArgs)
        {
            bool matched = false;

            foreach (Button_Koma[] kGroup in kGroupArgs)
            {
                foreach (Button_Koma btnKoma1 in kGroup)
                {
                    if (null != btnKoma1 && this.Sengo == btnKoma1.Masu.Sengo && this.Syurui == btnKoma1.Masu.Syurui)
                    {
                        // １つでも一致するものがあれば、終了します。
                        matched = true;
                        goto gt_EndLoop;
                    }
                }

            }
        gt_EndLoop:

            return matched;
        }

        /// <summary>
        /// 一致判定：　先後、駒種類  が、自分と同じもの。なければヌル。
        /// </summary>
        /// <returns></returns>
        public Button_Koma MatchFirst(params Button_Koma[][] kGroupArgs)
        {
            Button_Koma matched = null;

            foreach (Button_Koma[] kGroup in kGroupArgs)
            {
                foreach (Button_Koma k1 in kGroup)
                {
                    if (null != k1 && this.Sengo == k1.Masu.Sengo && this.Syurui == k1.Masu.Syurui)
                    {
                        // １つでも一致するものがあれば、終了します。
                        matched = k1;
                        goto gt_EndLoop;
                    }
                }

            }
        gt_EndLoop:

            return matched;
        }

        /// <summary>
        /// 一致判定：   場所指定
        /// </summary>
        /// <returns></returns>
        public bool Contains(MasuBango masu)
        {
            bool matched = false;

            if (masu.Match(this))
            {
                matched = true;
            }

            return matched;
        }

        /// <summary>
        /// 相手陣に入っていれば真。
        /// </summary>
        /// <returns></returns>
        public bool InAitejin()
        {
            return
                // 後手は 7,8,9 段。
                (Sengo.GOTE == this.Sengo && 7 <= this.Dan)
                // 先手は 1,2,3 段。
                || (Sengo.SENTE == this.Sengo && this.Dan <= 3);
        }

    }

}
